from typing import Dict, Optional, Any
import copy

# Type aliases
Object = Any
String = str


def property_of_object_if_exist(obj: Optional[Object], property_name: String) -> Optional[Any]:
    if obj is None:
        return None
    if hasattr(obj, "__getattribute__") and hasattr(object, property_name):
        return obj.__getattribute__(property_name)
    else:
        return None


def dictionary_of_object_if_exist(obj: Optional[Object]) -> Optional[Dict[Any, Any]]:
    return property_of_object_if_exist(obj, property_name="__dict__")


# Shallow purification from None on arbitrary Python Dictionaries. Returns purified dictionary.
def dictionary_purified_from_none(dictionary: Dict[Any, Any]) -> Dict[Any, Any]:
    return dict(filter(lambda item: item[1] is not None, dictionary.items()))


# Shallow purification from None on arbitrary Python Objects. Returns purified deepcopy of object.
def object_purified_from_none(obj: Object) -> Object:
    new_object = copy.deepcopy(obj)
    object_dictionary = dictionary_of_object_if_exist(new_object)
    if object_dictionary is not None:
        new_object.__dict__ = dictionary_purified_from_none(object_dictionary)
    return new_object



from .form_update import FormUpdate


class FormSubmitted(FormUpdate):
    def __init__(self, json_object):
        super().__init__(json_object)
        self.metadata = json_object.get("metadata")
        self.additional_metadata = json_object.get("additionalMetadata")



from typing import Optional, Union, List, Dict, Any
from pyAitu.utils.dictionary_extractor import dictionary_purified_from_none, dictionary_of_object_if_exist
from .options import Options
from .header import Header
from .bottom_bar import BottomBar
import logging

# Type aliases
Void = None
Array = List
Object = Any
String = str
Dictionaries = List[Dict[String, Object]]
ArrayOrObject = Union[Array[Object], Object]


# Function with side effect on input dictionaries
def add_purified_object_dictionary_if_possible_to(dictionaries: Dictionaries, obj: Object) -> Void:
    object_dictionary = dictionary_of_object_if_exist(obj)
    if object_dictionary is not None:
        dictionaries.append(dictionary_purified_from_none(object_dictionary))


def make_dictionaries_from(array_or_object: ArrayOrObject) -> Dictionaries:
    dictionaries: Dictionaries = []
    if isinstance(array_or_object, Array):
        array = array_or_object
        for obj in array:
            add_purified_object_dictionary_if_possible_to(dictionaries, obj)
    else:
        obj = array_or_object
        add_purified_object_dictionary_if_possible_to(dictionaries, obj)
    return dictionaries


class Form:
    def __init__(
            self,
            _id: String,
            content: ArrayOrObject,
            header: Header,
            options: Optional[Options] = None,
            bottom_bar: Optional[BottomBar] = None
    ):
        self.id = _id
        self.header = header
        self.content = content
        self.options = options
        self.bottom_bar = bottom_bar

    def make_dictionary(self) -> Dict[String, Any]:
        dictionary = {
            "id": self.id,
            "content": make_dictionaries_from(self.content)
        }
        header_dictionary = dictionary_of_object_if_exist(self.header)
        if header_dictionary:
            dictionary["header"] = header_dictionary
        else:
            logger = logging.getLogger('Form(make_dictionary)')
            logger.error('Can not get Dictionary of Header! Header is required attribute')
            return None

        if self.options:
            dictionary["options"] = dictionary_of_object_if_exist(self.options)

        if self.bottom_bar:
            dictionary["bottom_bar"] = dictionary_of_object_if_exist(self.bottom_bar)
        return { "form": dictionary }



from typing import List

from ..options import Options
from ..form_action import FormAction
from .file_metadata import FileMetadata
from .item_button import ItemButton


class Item:
    def __init__(
            self,
            item_id: str,
            title: str = None,
            subtitle: str = None,
            options: Options = None,
            form_action: FormAction = None,
            file_metadata: FileMetadata = None,
            item_buttons: List[ItemButton] = None
    ):
        self.id = item_id
        self.title = title
        self.subtitle = subtitle
        if options is not None:
            self.options = options.__dict__
        if form_action is not None:
            self.form_action = form_action.__dict__
        if file_metadata is not None:
            self.file_metadata = file_metadata.__dict__
        if item_buttons is not None:
            self.item_buttons = item_buttons
        self.__dict__ = self._remove_none()

    def _remove_none(self):
        return {k:v for k,v in self.__dict__.items() if v is not None}



class FlexOptions:
    def __init__(
            self,
            flex_grow: float = None,
            flex_basis: float = None,
            flex_direction: str = None,
            flex_wrap: str = None,
            align_items: str = None,
            align_self: str = None,
            justify_content: str = None):
        self.flex_grow = flex_grow
        self.flex_basis = flex_basis
        self.flex_direction = flex_direction
        self.flex_wrap = flex_wrap
        self.align_items = align_items
        self.align_self = align_self
        self.justify_content = justify_content



import inspect
import re
from ..utils import context
from ..models import ContentType, Message

USER_STATE = 'USER_STATE'


async def check_filter(filter_, args):
    if not callable(filter_):
        raise TypeError('Filter must be callable and/or awaitable!')

    if inspect.isawaitable(filter_) or inspect.iscoroutinefunction(filter_):
        return await filter_(*args)
    else:
        return filter_(*args)


async def check_filters(filters, args):
    if filters is not None:
        for filter_ in filters:
            f = await check_filter(filter_, args)
            if not f:
                return False
    return True


def generate_default_filters(dispatcher, *args, **kwargs):
    filter_set = []

    for name, filter_ in kwargs.items():
        if filter_ is None and name != DefaultFilters.STATE:
            continue
        if name == DefaultFilters.COMMANDS:
            if isinstance(filter_, str):
                filter_set.append(CommandsFilter([filter_]))
            else:
                filter_set.append(CommandsFilter(filter_))
        elif name == DefaultFilters.CONTENT_TYPES:
            filter_set.append(ContentTypeFilter(filter_))
        elif name == DefaultFilters.REGEXP:
            filter_set.append(RegexpFilter(filter_))
        elif name == DefaultFilters.STATE:
            if isinstance(filter_, (list, set, tuple)):
                filter_set.append(StatesListFilter(dispatcher, filter_))
            else:
                filter_set.append(StateFilter(dispatcher, filter_))
        elif isinstance(filter_, Filter):
            filter_set.append(filter_)

    filter_set += list(args)

    return filter_set


class Filter:
    def __call__(self, *args, **kwargs):
        return self.check(*args, **kwargs)

    def check(self, *args, **kwargs):
        raise NotImplementedError


class AsyncFilter(Filter):
    def __aiter__(self):
        return None

    def __await__(self):
        return self.check

    async def check(self, *args, **kwargs):
        pass


class CommandsFilter(AsyncFilter):
    def __init__(self, commands):
        self.commands = commands

    async def check(self, message):
        if not message.is_command():
            return False

        command = message.text.split()[0][1:]
        command, _, mention = command.partition('@')

        if mention and mention != (await message.bot.me).username:
            return False

        if command not in self.commands:
            return False

        return True


class ContentTypeFilter(Filter):
    def __init__(self, content_types):
        self.content_types = content_types

    def check(self, message):
        return ContentType.ANY[0] in self.content_types or message.content_type in self.content_types


class RegexpFilter(Filter):
    def __init__(self, regexp):
        self.regexp = re.compile(regexp, flags=re.IGNORECASE | re.MULTILINE)

    def check(self, _object):
        if isinstance(_object, Message) and _object.text:
            return bool(self.regexp.search(_object.text))


class StateFilter(AsyncFilter):
    def __init__(self, dispatcher, state):
        self.dispatcher = dispatcher
        self.state = state

    def get_target(self, object_):
        return getattr(getattr(object_, 'chat', None), 'id', None), \
               getattr(getattr(object_, 'from_user', None), 'id', None)

    async def check(self, object_):
        if self.state == '*':
            return True

        if context.check_value(USER_STATE):
            context_state = context.get_value(USER_STATE)
            return self.state == context_state
        else:
            chat, user = self.get_target(object_)

            if chat or user:
                return await self.dispatcher.storage.get_state(chat=chat, user=user) == self.state
        return False


class StatesListFilter(StateFilter):
    async def check(self, object_):
        chat, user = self.get_target(object_)

        if chat or user:
            return await self.dispatcher.storage.get_state(chat=chat, user=user) in self.state
        return False


class DefaultFilters:
    COMMANDS = "commands"
    REGEXP = "regexp"
    CONTENT_TYPES = "content_types"
    FUNC = "func"
    STATE = "state"



from pyAitu.models.form.content.content import Content
from pyAitu.models.form.form_action import FormAction
from pyAitu.utils.dictionary_extractor import dictionary_of_object_if_exist


class BottomBar(Content):

    def __init__(self, content_id: str, title: str, form_action: FormAction):
        super().__init__("bottom_bar", content_id)
        self.title = title
        self.form_action = dictionary_of_object_if_exist(form_action)



from typing import List, Dict, Optional
from .base import BaseBot
from ..models import Update, Media, Command, QuickButtonCommand, InlineCommand, ReplyCommand, Form, Contact, \
    WebhookInfo, SetWebhook, FileType
from ..utils.strings import COMMANDS, SEND_MESSAGE, GET_UPDATES, UPLOADED_FILES, SEND_UI_STATE, SEND_CONTACT_MESSAGE, \
    EDIT_MESSAGE, SEND_CONTAINER_MESSAGE, FORWARD_MESSAGE
import json


class Bot(BaseBot):
    async def get_updates(self) -> List[Update]:
        result = await self.request(method=GET_UPDATES)
        return [Update(update) for update in result.get("updates", [])]

    async def send_message(self,
                           peer_id: str,
                           content: str,
                           quick_button_commands: List[QuickButtonCommand] = None,
                           inline_commands: List[InlineCommand] = None,
                           reply_keyboard: List[ReplyCommand] = None,
                           media_list: List[Media] = None,
                           local_id: str = None
                           ) -> Dict:
        command = Command(inline_commands=inline_commands, media=media_list)

        payload = {
            COMMANDS: command.create_command(
                SEND_MESSAGE,
                peer_id = peer_id,
                content=content,
                reply_keyboard=reply_keyboard,
                quick_button_commands=quick_button_commands,
                local_id=local_id
            )
        }

        result = await self.request(SEND_MESSAGE, payload)
        return result

    async def forward_message(self,
                              from_dialog,
                              to_dialog: str,
                              message_id: str,
                              local_id: str = None
                              ):
        command = Command()

        payload = {
            COMMANDS: command.create_command(
                FORWARD_MESSAGE,
                from_dialog=from_dialog,
                to_dialog=to_dialog,
                message_id=message_id,
                local_id=local_id
            )
        }

        result = await self.request(SEND_MESSAGE, payload)
        return result

    async def edit_message(self,
                           peer_id: str,
                           message_id: str,
                           content: str,
                           inline_commands: List[InlineCommand] = None
                           ):
        command = Command(inline_commands=inline_commands)

        payload = {
            COMMANDS: command.create_command(
                EDIT_MESSAGE,
                peer_id=peer_id,
                content=content,
                message_id=message_id
            )
        }

        result = await self.request(EDIT_MESSAGE, payload)
        return result

    async def send_media(self,
                         chat_id: str,
                         file: str,
                         file_type: FileType):
        result = await self.upload_file(file)
        if result.get(UPLOADED_FILES):
            media = Media(
                file_id=result.get(UPLOADED_FILES)[0]["fileId"],
                name=result.get(UPLOADED_FILES)[0]["fileName"],
                file_type=file_type
            )
            command = Command(media=[media])

            payload = {
                COMMANDS: command.create_command(SEND_MESSAGE, chat_id)
            }
            result = await self.request(SEND_MESSAGE, payload)

            return result

    async def send_form(
            self,
            chat_id: str,
            form: Form
    ):
        command = Command()
        payload = {
            COMMANDS: command.create_command(SEND_UI_STATE, chat_id, form=form)
        }

        result = await self.request(SEND_UI_STATE, payload)
        return result

    async def send_container_message(
            self,
            chat_id: str,
            content: List
    ):

        dict_content = []

        for i in content:
            dict_content.append(i.__dict__)

        str_content = json.dumps(dict_content)

        command = Command()
        payload = {
            COMMANDS: command.create_command(SEND_CONTAINER_MESSAGE, chat_id, content=str_content)
        }

        result = await self.request(SEND_CONTAINER_MESSAGE, payload)
        return result

    async def upload_file(self, file):
        files = {
            "file": file
        }

        return await self.request("UploadFile", None, files)

    async def send_contact(
            self,
            chat_id: str,
            contact: Contact
    ):
        command = Command()
        payload = {
            COMMANDS: command.create_command(SEND_CONTACT_MESSAGE, chat_id, input_media=contact)
        }

        result = await self.request(SEND_UI_STATE, payload)
        return result

    async def get_webhook(self) -> WebhookInfo:
        result = await self.request("GetWebhook")
        return WebhookInfo(result)

    async def set_webhook(self, url: str):
        result = await self.request("SetWebhook", SetWebhook(url).__dict__)
        return result

    async def delete_webhook(self):
        return await self.request("DeleteWebhook")



from pyAitu.models.form.content import Button, Submit
from .content import SimpleCatalog, Item, LabeledText, TextArea, ItemInfo, SimpleCatalog, Item, Submit, Divider, Image,\
    FileMetadata, DatePicker, Text, Input, UserInfo, Checkbox, Switch, Radiogroup, CustomContainer, MediaPicker
from .options import Options
from .form import Form
from .header import Header
from .validation_rule import ValidationRule
from .flex_options import FlexOptions
from .form_action import FormAction
from .indent import Indent
from ..constants import InputType, Currency
from .bottom_bar import BottomBar

__all__ = [
    SimpleCatalog,
    Item,
    ItemInfo,
    TextArea,
    Options,
    Form,
    Header,
    ValidationRule,
    LabeledText,
    FlexOptions,
    Button,
    Submit,
    FormAction,
    Divider,
    Image,
    DatePicker,
    MediaPicker,
    FileMetadata,
    Switch,
    Text,
    Indent,
    Input,
    InputType,
    Currency,
    BottomBar,
    Radiogroup,
    CustomContainer,
    Checkbox,
    UserInfo
]



class FileMetadata:
    def __init__(
            self,
            file_type: str,
            file_id: str,
            file_name: str
    ):
        self.file_type = file_type
        self.file_id = file_id
        self.filename = file_name



import asyncio
from typing import Dict, Coroutine

CONFIGURED = '@CONFIGURED_TASK_FACTORY'


def task_factory(loop: asyncio.BaseEventLoop, coroutine: Coroutine):
    if loop.is_closed():
        raise RuntimeError('Event loop is closed')

    task = asyncio.Task(coroutine, loop=loop)

    try:
        task.context = asyncio.Task.current_task().context.copy()
    except AttributeError:
        task.context = {CONFIGURED: True}

    return task


def set_value(key, value):
    get_current_state()[key] = value


def get_current_state() -> Dict:
    task = asyncio.Task.current_task()
    if task is None:
        raise RuntimeError('Can be used only in Task context.')
    context_ = getattr(task, 'context', None)
    if context_ is None:
        context_ = task.context = {}
    return context_


def update_state(data=None, **kwargs):
    if data is None:
        data = {}
    state = get_current_state()
    state.update(data, **kwargs)


def check_value(key):
    return key in get_current_state()


def get_value(key, default=None):
    return get_current_state().get(key, default)



from typing import Union, Optional


class BaseStorage:
    async def get_state(self, *,
                        chat: Union[str, int, None] = None,
                        user: Union[str, int, None] = None,
                        default: Optional[str] = None) -> Optional[str]:
        raise NotImplementedError


class DisabledStorage(BaseStorage):
    async def get_state(self, *,
                        chat: Union[str, int, None] = None,
                        user: Union[str, int, None] = None,
                        default: Optional[str] = None) -> Optional[str]:
        return None


from ..options import Options
from .content import Content


class Divider(Content):
    def __init__(self, content_id, options: Options = None):
        super().__init__(content_type="divider", content_id=content_id)
        if options is not None:
            self.options = options.__dict__



from ...utils.strings import SHOW_SHARE_CONTACT_BUTTON, SHOW_RECORD_AUDIO_BUTTON, SHOW_GALLERY_BUTTON, \
    SHOW_CAMERA_BUTTON, REPLY_KEYBOARD, QUICK_BUTTON_COMMANDS, FORM_MESSAGE


class UiState:
    def __init__(self,
                 show_camera_button=True,
                 show_share_contact_button=True,
                 show_record_audio_button=True,
                 show_gallery_button=True,
                 can_write_text=False,
                 reply_keyboard: list = None,
                 quick_button_commands: list = None,
                 form_message: dict = None
                 ):
        self.show_camera_button = show_camera_button
        self.show_share_contact_button = show_share_contact_button
        self.show_record_audio_button = show_record_audio_button
        self.show_gallery_button = show_gallery_button
        self.can_write_text = can_write_text
        self.form_message = form_message
        self.reply_keyboard = []
        self.quick_button_commands = []

        if reply_keyboard:
            for command in reply_keyboard:
                self.reply_keyboard.append(command.to_dict())

        if quick_button_commands:
            for command in quick_button_commands:
                self.quick_button_commands.append(command.to_dict())

    def to_dict(self):
        return {
            SHOW_CAMERA_BUTTON: self.show_camera_button,
            SHOW_GALLERY_BUTTON: self.show_gallery_button,
            SHOW_RECORD_AUDIO_BUTTON: self.show_record_audio_button,
            SHOW_SHARE_CONTACT_BUTTON: self.show_share_contact_button,
            REPLY_KEYBOARD: self.reply_keyboard,
            QUICK_BUTTON_COMMANDS: self.quick_button_commands,
            FORM_MESSAGE: self.form_message
        }



import asyncio
import aiohttp
import json
import io
from typing import Optional, Union, List, Dict
from . import api


class BaseBot:
    def __init__(
            self,
            token: str,
            loop: Optional[Union[asyncio.BaseEventLoop, asyncio.AbstractEventLoop]] = None
    ):
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        self.token = token
        connector = aiohttp.TCPConnector(loop=self.loop, verify_ssl=False)
        self.session = aiohttp.ClientSession(connector=connector, loop=self.loop, json_serialize=json.dumps)

    async def request(self,
                      method: str,
                      data: Optional[Dict] = None,
                      files: Optional[Dict] = None) -> Union[List, Dict, bool]:
        return await api.request(session=self.session, token=self.token, method=method, data=data, files=files)

    async def download_file(
            self,
            file_id: str,
            destination=None,
            chunk_size: int = 65536,
            seek: bool = True
    ):
        if destination is None:
            destination = io.BytesIO()

        url = "{}?fileId={}".format(api.FILE_DOWNLOAD_URL, file_id)
        destination_ = destination if isinstance(destination, io.IOBase) else open(destination, 'wb')
        async with self.session.get(url, headers={"X-BOT-TOKEN": self.token}) as response:
            while True:
                chunk = await response.content.read(chunk_size)
                if not chunk:
                    break
                destination_.write(chunk)
                destination_.flush()
        if seek:
            destination_.seek(0)
        return destination_



from .filters import check_filters
from ..utils import context


class SkipHandler(BaseException):
    pass


class CancelHandler(BaseException):
    pass


class Handler:
    def __init__(self, dispatcher, once=True, middleware_key=None):
        self.dispatcher = dispatcher
        self.middleware_key = middleware_key
        self.once = once
        self.handlers = []

    def register(self, handler, filters=None, index=None):
        if filters and not isinstance(filters, (list, tuple, set)):
            filters = [filters]
        record = (filters, handler)
        if index is None:
            self.handlers.append(record)
        else:
            self.handlers.insert(index, record)

    def unregister(self, handler):
        for handler_with_filters in self.handlers:
            _, registered = handler_with_filters
            if handler is registered:
                self.handlers.remove(handler_with_filters)
                return True
        raise ValueError('This handler is not registered')

    async def notify(self, *args):
        results = []

        if self.middleware_key:
            try:
                await self.dispatcher.middleware.trigger(f"pre_process_{self.middleware_key}", args)
            except CancelHandler:
                return results

        for filters, handler in self.handlers:
            if await check_filters(filters, args):
                try:
                    if self.middleware_key:
                        context.set_value('handler', handler)
                        await self.dispatcher.middleware.trigger(f"process_{self.middleware_key}", args)
                    response = await handler(*args)
                    if response is not None:
                        results.append(response)
                    if self.once:
                        break
                except SkipHandler:
                    continue
                except CancelHandler:
                    break

        if self.middleware_key:
            await self.dispatcher.middleware.trigger(f"post_process_{self.middleware_key}", args + (results,))

        return results



from .update import Update, QuickButtonSelected, InlineCommandSelected, Message, ContentType, FormClosed, \
    FormSubmitted, FormMessageSent, MessageIdAssigned
from .media import Media, Contact, Audio, FileType
from .webhook import WebhookInfo, SetWebhook
from .command import InlineCommand, ReplyCommand, QuickButtonCommand, Command, UiState
from .form import SimpleCatalog, Item, ItemInfo, Options, Form, Header, TextArea, ValidationRule, FlexOptions, Button, \
    FormAction, Submit, LabeledText, Divider, Image, FileMetadata, Text, Indent, Input, Currency, UserInfo,\
    DatePicker, Checkbox, Switch, Radiogroup, CustomContainer, BottomBar, MediaPicker
from .constants import Alignment, Orientation, Currency, FileType, TextSize, TextStyle, OptionMediaType, InputType

__all__ = [
    Message,
    ContentType,
    Update,
    QuickButtonSelected,
    InlineCommandSelected,
    Media,
    Contact,
    Audio,
    InlineCommand,
    ReplyCommand,
    QuickButtonCommand,
    Command,
    UiState,
    SimpleCatalog,
    Item,
    ItemInfo,
    Options,
    Form,
    Header,
    FormClosed,
    TextArea,
    LabeledText,
    ValidationRule,
    FlexOptions,
    Button,
    Alignment,
    Orientation,
    Currency,
    FileType,
    TextStyle,
    TextSize,
    OptionMediaType,
    InputType,
    FormSubmitted,
    Submit,
    FormAction,
    FormMessageSent,
    MessageIdAssigned,
    Divider,
    Image,
    DatePicker,
    MediaPicker,
    FileMetadata,
    Switch,
    Text,
    Indent,
    Input,
    Currency,
    BottomBar,
    Radiogroup,
    CustomContainer,
    Checkbox,
    UserInfo,
    WebhookInfo,
    SetWebhook,
]



from .form_update import FormUpdate


class FormMessageSent(FormUpdate):
    def __init__(self, json_object):
        super().__init__(json_object)
        self.message = json_object.get("message")
        self.message_id = json_object.get("messageId")
        self.additional_metadata = json_object.get("additionalMetadata")



from ..peer import Peer


class FormUpdate:
    def __init__(self, json_object):
        self.update_id = json_object.get("updateId")
        self.form_id = json_object.get("formId")
        self.dialog = Peer(json_object.get("dialog"))
        self.sender = Peer(json_object.get("sender"))

    @property
    def chat(self):
        return self.dialog



from .alignment import Alignment
from .currency import Currency
from .file_type import FileType
from .orientation import Orientation
from .text_size import TextSize
from .text_style import TextStyle
from .option_media_type import OptionMediaType
from .input_type import InputType

__all__ = [
    Alignment,
    Currency,
    FileType,
    Orientation,
    TextStyle,
    TextSize,
    OptionMediaType,
    InputType,
]


from ...utils.strings import CAPTION, METADATA


class InlineCommand:
    def __init__(self, caption: str, metadata: str):
        self._caption = caption
        self._metadata = metadata

    @property
    def caption(self):
        return self._caption

    @property
    def metadata(self):
        return self._metadata

    def to_dict(self):
        return {
            CAPTION: self.caption,
            METADATA: self.metadata
        }



from typing import Optional, List

from ..options import Options
from ..validation_rule import ValidationRule
from .content import Content
from ....utils.serializer import serialized

ValidationRules = List[ValidationRule]


class Checkbox(Content):
    def __init__(
            self,
            content_id: str,
            title: str,
            default_state: bool = False,
            options: Optional[Options] = None,
            validations_rules: Optional[ValidationRules] = None,
    ):
        super().__init__(content_id=content_id, content_type="checkbox")
        self.title = title
        self.default_state = default_state
        if options is not None:
            self.options = options.__dict__
        if validations_rules is not None:
            self.validations_rules = serialized(validations_rules)



from ..peer import Peer


class InlineCommandSelected:
    def __init__(self, json_object):
        self.update_id = json_object.get("updateId")
        self.message_id = json_object.get("messageId")
        self.dialog = Peer(json_object.get("dialog"))
        self.sender = Peer(json_object.get("sender"))
        self.metadata = json_object.get("metadata")
        self.content = json_object.get("content")

    @property
    def from_user(self):
        return self.sender



from typing import List, Optional

from ..options import Options
from pyAitu.utils.serializer import serialized
from ..validation_rule import ValidationRule
from .content import Content

ValidationRules = List[ValidationRule]


class DatePicker(Content):
    def __init__(
            self,
            content_id: str,
            validations_rules: Optional[ValidationRules] = None,
            title: Optional[str] = None,
            selected_date: Optional[str] = None,
            options: Optional[Options] = None
    ):
        super().__init__(content_type="date_picker", content_id=content_id)
        self.title = title
        self.selected_date = selected_date
        if options is not None:
            self.options = options.__dict__
        self.validations_rules = serialized(validations_rules)



from ..options import Options
from .content import Content


class Slider(Content):
    def __init__(self, content_id: str, items: list, options: Options = None):
        super().__init__(content_type="slider", content_id=content_id)
        self.items = self.serialize_items(items)
        if options is not None:
            self.options = options.__dict__

    def serialize_items(self, items: list):
        result = []
        for item in items:
            result.append(item.__dict__)
        return result



from ..form_action import FormAction


class ItemButton:
    def __init__(
            self,
            title: str,
            form_action: FormAction
    ):
        self.title = title
        self.form_action = form_action



from .content import Content
from ..options import Options
from ..form_action import FormAction


class Button(Content):
    def __init__(
            self,
            content_id: str,
            title: str = None,
            button_type: str = None,
            options: Options = None,
            form_action: FormAction = None
    ):
        super().__init__(content_type="button", content_id=content_id)
        self.title = title
        self.button_type = button_type
        self.options = options.__dict__
        self.form_action = form_action.__dict__



from typing import Optional
from ..constants import Currency, Alignment, Orientation, TextSize, TextStyle, OptionMediaType, InputType
from .indent import Indent
from .flex_options import FlexOptions
from ...utils.dictionary_extractor import dictionary_purified_from_none, dictionary_of_object_if_exist


class Options:
    def __init__(
            self,
            type: Optional[str] = None,
            alignment: Optional[Alignment] = None,
            background: Optional[str] = None,
            background_color: Optional[str] = None,
            choice_type: Optional[str] = None,
            closeable: Optional[bool] = None,
            columns_count: Optional[int] = None,
            currency: Optional[Currency] = None,
            divider_type: Optional[str] = None,
            flex_options: Optional[FlexOptions] = None,
            fullscreen: Optional[bool] = None,
            has_back_action: Optional[bool] = None,
            height: Optional[int] = None,
            indent_inner: Optional[Indent] = None,
            indent_outer: Optional[Indent] = None,
            input_type: Optional[InputType] = None,
            item_left_icon_resource: Optional[str] = None,
            item_right_icon_resource: Optional[str] = None,
            item_type: Optional[str] = None,
            max_count: Optional[int] = None,
            max_date: Optional[str] = None,
            max_length: Optional[int] = None,
            media_type: Optional[OptionMediaType] = None,
            min_date: Optional[str] = None,
            orientation: Optional[Orientation] = None,
            search_enabled: Optional[bool] = None,
            show_divider: Optional[bool] = None,
            shape: Optional[str] = None,
            should_open_editor: Optional[bool] = None,
            subtitle_lines_count: Optional[int] = None,
            text_color: Optional[str] = None,
            text_size: Optional[TextSize] = None,
            text_style: Optional[TextStyle] = None,
            title: Optional[str] = None,
            title_lines_count: Optional[int] = None,
            width: Optional[int] = None
    ):
        self.type = type
        self.alignment = alignment
        self.background = background
        self.background_color = background_color
        self.choice_type = choice_type
        self.closeable = closeable
        self.columns_count = columns_count
        self.currency = currency
        self.divider_type = divider_type
        if flex_options is not None:
            self.flex_options = flex_options.__dict__
        self.fullscreen = fullscreen
        self.has_back_action = has_back_action
        self.height = height
        if indent_inner is not None:
            self.indent_inner = indent_inner.__dict__
        if indent_outer is not None:
            self.indent_outer = indent_outer.__dict__
        self.input_type = input_type
        self.item_left_icon_resource = item_left_icon_resource
        self.item_right_icon_resource = item_right_icon_resource
        self.item_type = item_type
        self.max_count = max_count
        self.max_date = max_date
        self.max_length = max_length
        self.media_type = media_type
        self.min_date = min_date
        self.orientation = orientation
        self.search_enabled = search_enabled
        self.show_divider = show_divider
        self.shape = shape
        self.should_open_editor = should_open_editor
        self.text_color = text_color
        self.title = title
        self.title_lines_count = title_lines_count
        self.subtitle_lines_count = subtitle_lines_count
        self.text_size = text_size
        self.text_style = text_style
        self.width = width
        # Clean None values up
        self.purify_none_attributes()

    # Removes all None attributes of self
    # Mutating function, with side effect
    def purify_none_attributes(self):
        options_dictionary = dictionary_of_object_if_exist(obj=self)
        if options_dictionary:
            self.__dict__ = dictionary_purified_from_none(options_dictionary)




from .message import Message
from .message_id_assigned import MessageIdAssigned
from .quick_button_selected import QuickButtonSelected
from .inline_command_selected import InlineCommandSelected
from .form_closed import FormClosed
from .form_submitted import FormSubmitted
from .form_message_sent import FormMessageSent


class Update:

    message: Message = None
    quick_button_selected: QuickButtonSelected = None
    inline_command_selected: InlineCommandSelected = None
    form_closed: FormClosed = None
    form_submitted: FormSubmitted = None
    form_message_sent: FormMessageSent = None
    message_id_assigned: MessageIdAssigned = None

    def __init__(self, json_object):
        self.updateId = json_object.get("updateId")
        self.updateType = json_object.get("type")

        if self.updateType == "Message":
            self.message = Message(json_object)
        if self.updateType == "QuickButtonSelected":
            self.quick_button_selected = QuickButtonSelected(json_object)
        if self.updateType == "InlineCommandSelected":
            self.inline_command_selected = InlineCommandSelected(json_object)
        if self.updateType == "FormMessageSent":
            self.form_message_sent = FormMessageSent(json_object)
        if self.updateType == "FormSubmitted":
            self.form_submitted = FormSubmitted(json_object)
        if self.updateType == "FormClosed":
            self.form_closed = FormClosed(json_object)
        if self.updateType == "MessageIdAssigned":
            self.message_id_assigned = MessageIdAssigned(json_object)



from ..peer import Peer


class MessageIdAssigned:
    def __init__(self, json_object):
        self.updateId = json_object.get("updateId")
        self.localId = json_object.get("localId")
        self.dialog = Peer(json_object.get("dialog"))
        self.id = json_object.get("id")

    @property
    def chat(self):
        return self.dialog



from ..options import Options
from .content import Content


class LabeledText(Content):
    def __init__(
            self,
            content_id: str,
            label: str,
            title: str,
            options: Options = None
    ):
        super().__init__(content_type="labeled_text", content_id=content_id)
        self.label = label
        self.title = title
        if options is not None:
            self.options = options.__dict__



from .simple_catalog import SimpleCatalog
from .text_area import TextArea
from .labeled_text import LabeledText
from .item import Item
from .item_info import ItemInfo
from .button import Button
from .submit import Submit
from .image import Image
from .date_picker import DatePicker
from .media_picker import MediaPicker
from .file_metadata import FileMetadata
from .input import Input
from .switch import Switch
from .divider import Divider
from .text import Text
from .radiogroup import Radiogroup
from .custom_container import CustomContainer
from .checkbox import Checkbox
from .user_info import UserInfo


__all__ = [
    SimpleCatalog,
    Submit,
    TextArea,
    LabeledText,
    Item,
    ItemInfo,
    Input,
    DatePicker,
    MediaPicker,
    Divider,
    Button,
    Image,
    Switch,
    FileMetadata,
    Text,
    Radiogroup,
    CustomContainer,
    Checkbox,
    UserInfo
]



class Contact():
    def __init__(self, first_name: str = None, last_name: str = None, phone_number: str = None):
        self.firstName = first_name
        self.lastName = last_name
        self.phoneNumber = phone_number
        self.contactType = "InputUnregisteredContact"



from ..options import Options
from .content import Content


class Switch(Content):
    def __init__(self, content_id: str, title: str = None, default_state: bool = False, options: Options = None):
        super().__init__(content_type="switch", content_id=content_id)
        self.title = title
        self.default_state = default_state
        if options is not None:
            self.options = options.__dict__



from .update import Update
from .quick_button_selected import QuickButtonSelected
from .inline_command_selected import InlineCommandSelected
from .message import Message, ContentType
from .form_closed import FormClosed
from .form_submitted import FormSubmitted
from .form_message_sent import FormMessageSent
from .message_id_assigned import MessageIdAssigned

__all__ = [
    Update,
    QuickButtonSelected,
    InlineCommandSelected,
    Message,
    ContentType,
    FormClosed,
    FormSubmitted,
    FormMessageSent,
    MessageIdAssigned,
]



class Helper:
    mode = ''

    @classmethod
    def all(cls):
        result = []

        for name in dir(cls):
            if not name.isupper():
                continue
            value = getattr(cls, name)
            if isinstance(value, ItemsList):
                result.append(value[0])
            else:
                result.append(value)
        return result


class HelperMode(Helper):
    SCREAMING_SNAKE_CASE = 'SCREAMING_SNAKE_CASE'
    lowerCamelCase = 'lowerCamelCase'
    CamelCase = 'CamelCase'
    snake_case = 'snake_case'
    lowercase = 'lowercase'

    @classmethod
    def all(cls):
        return [
            cls.SCREAMING_SNAKE_CASE,
            cls.lowerCamelCase,
            cls.CamelCase,
            cls.snake_case,
            cls.lowercase,
        ]

    @classmethod
    def _screaming_snake_case(cls, text):
        if text.isupper():
            return text
        result = ''
        for pos, symbol in enumerate(text):
            if symbol.isupper() and pos > 0:
                result += '_' + symbol
            else:
                result += symbol.upper()
        return result

    @classmethod
    def _snake_case(cls, text):
        if text.islower():
            return text
        return cls._screaming_snake_case(text).lower()

    @classmethod
    def _camel_case(cls, text, first_upper=False):
        result = ''
        need_upper = False
        for pos, symbol in enumerate(text):
            if symbol == '_' and pos > 0:
                need_upper = True
            else:
                if need_upper:
                    result += symbol.upper()
                else:
                    result += symbol.lower()
                need_upper = False
        if first_upper:
            result = result[0].upper() + result[1:]
        return result

    @classmethod
    def apply(cls, text, mode):
        if mode == cls.SCREAMING_SNAKE_CASE:
            return cls._screaming_snake_case(text)
        if mode == cls.snake_case:
            return cls._snake_case(text)
        if mode == cls.lowercase:
            return cls._snake_case(text).replace('_', '')
        if mode == cls.lowerCamelCase:
            return cls._camel_case(text)
        if mode == cls.CamelCase:
            return cls._camel_case(text, True)
        if callable(mode):
            return mode(text)
        return text


class Item:
    def __init__(self, value=None):
        self._value = value

    def __get__(self, instance, owner):
        return self._value

    def __set_name__(self, owner, name):
        if not name.isupper():
            raise NameError('Name for helper item must be in uppercase!')
        if not self._value:
            if hasattr(owner, 'mode'):
                self._value = HelperMode.apply(name, getattr(owner, 'mode'))


class ListItem(Item):
    def add(self, other):
        return self + other

    def __get__(self, instance, owner):
        return ItemsList(self._value)

    def __getitem__(self, item):
        return self._value

    __iadd__ = __add__ = __rand__ = __and__ = __ror__ = __or__ = add


class ItemsList(list):
    def __init__(self, *seq):
        super(ItemsList, self).__init__(map(str, seq))

    def add(self, other):
        self.extend(other)
        return self

    __iadd__ = __add__ = __rand__ = __and__ = __ror__ = __or__ = add



from .media import Media
from .image import Image
from .audio import Audio
from .contact import Contact
from pyAitu.models.constants.file_type import FileType

__all__ = [
    FileType,
    Media,
    Image,
    Audio,
    Contact,
]



from .peer import Peer


class PeerUser(Peer):
    def __init__(self, json_object):
        super().__init__(json_object)
        self._first_name = json_object['firstName'] if 'firstName' in json_object else None
        self._last_name = json_object['lastName'] if 'lastName' in json_object else None
        self._username = json_object['username'] if 'username' in json_object else None



from ...utils.strings import CAPTION, METADATA, ACTION


class QuickButtonCommand:
    def __init__(self, caption: str, metadata: str, action: str = 'QUICK_REQUEST'):
        self._caption = caption
        self._metadata = metadata
        self._action = action

    @property
    def caption(self):
        return self._caption

    @property
    def metadata(self):
        return self._metadata

    @property
    def action(self):
        return self._action

    def to_dict(self):
        return {
            CAPTION: self.caption,
            METADATA: self.metadata,
            ACTION: self.action
        }



DEFAULT_USER = "USER"
TYPE = "type"
ID = "id"
UI_STATE = "uiState"
RECIPIENT = "recipient"
CONTENT = "content"
COMMANDS = "commands"
CAPTION = "caption"
METADATA = "metadata"
ACTION = "action"
FILE_ID = "fileId"
NAME = "name"
FILE_TYPE = "mediaType"
SEND_MESSAGE = "SendMessage"
FORWARD_MESSAGE = "ForwardMessage"
EDIT_MESSAGE = "EditMessage"
SEND_UI_STATE = "SendUiState"
SEND_CONTAINER_MESSAGE = "SendContainerMessage"
SEND_CONTACT_MESSAGE = "SendContactMessage"
MEDIA_LIST = "mediaList"
LOCAL_ID = "localId"
GET_UPDATES = "GetUpdates"
UPLOADED_FILES = "uploadedFiles"
QUICK_BUTTON_COMMANDS = "quickButtonCommands"
FORM_MESSAGE = "formMessage"
INLINE_COMMANDS = "inlineCommands"
REPLY_KEYBOARD = "replyKeyboard"
SHOW_CAMERA_BUTTON = "showCameraButton"
SHOW_SHARE_CONTACT_BUTTON = "showShareContactButton"
SHOW_RECORD_AUDIO_BUTTON = "showRecordAudioButton"
SHOW_GALLERY_BUTTON = "showGalleryButton"
INPUT_MEDIA = "inputMedia"
MESSAGE_ID = "messageId"
FROM_DIALOG = "fromDialog"
TO_DIALOG = "toDialog"



from ..options import Options
from .content import Content
from pyAitu.models.form.form_action import FormAction
from pyAitu.utils.dictionary_extractor import dictionary_of_object_if_exist


class Text(Content):
    def __init__(self, content_id, title: str, options: Options = None, form_action: FormAction = None):
        super().__init__(content_type="text", content_id=content_id)
        self.title = title
        self.form_action = dictionary_of_object_if_exist(form_action)
        if options is not None:
            self.options = options.__dict__



from ..options import Options
from .content import Content
from .file_metadata import FileMetadata


class ItemInfo(Content):
    def __init__(
            self,
            content_id: str,
            title: str = None,
            subtitle: str = None,
            file_metadata: FileMetadata = None,
            options: Options = None
    ):
        super().__init__(content_type="item_info", content_id=content_id)
        self.title = title
        self.subtitle = subtitle
        if file_metadata is not None:
            self.file_metadata = file_metadata.__dict__
        if options is not None:
            self.options = options.__dict__



from typing import List
from .uiState import UiState
from .form_message import FormMessage
from ..peer import Recipient
from ..form import Form
from ..media import Media, Contact
from ...utils.strings import UI_STATE, RECIPIENT, TYPE, CONTENT, INLINE_COMMANDS, MEDIA_LIST, INPUT_MEDIA, LOCAL_ID, \
    MESSAGE_ID, FROM_DIALOG, TO_DIALOG


class Command:
    def __init__(self, inline_commands: list = None, media: List[Media] = None):
        self.media = []
        self.inline_commands = []

        if inline_commands:
            for command in inline_commands:
                self.inline_commands.append(command.to_dict())

        if media:
            for item in media:
                self.media.append(item.to_dict())

    def create_command(
            self,
            _type: str,
            peer_id: str = None,
            content: str = "",
            reply_keyboard: list = None,
            quick_button_commands: list = None,
            form: Form = None,
            input_media: Contact = None,
            local_id: str = None,
            message_id: str = None,
            from_dialog: str = None,
            to_dialog: str = None
    ):
        commands = []
        form_message = FormMessage(form).__dict__ if form is not None else {}
        if input_media is not None:
            input_media = input_media.__dict__
        ui_state = UiState(
            reply_keyboard=reply_keyboard,
            quick_button_commands=quick_button_commands,
            form_message=form_message
        )

        body = {
            TYPE: _type,
            RECIPIENT: Recipient(peer_id).get_recipient(),
            CONTENT: content,
            UI_STATE: ui_state.to_dict(),
            INLINE_COMMANDS: self.inline_commands,
            MEDIA_LIST: self.media,
            INPUT_MEDIA: input_media,
            LOCAL_ID: local_id,
            MESSAGE_ID: message_id,
            FROM_DIALOG: Recipient(from_dialog).get_recipient(),
            TO_DIALOG: Recipient(to_dialog).get_recipient()
         }

        commands.append(self.remove_none(body))

        return commands

    def remove_none(self, command: dict):
        return {k: v for k, v in command.items() if v is not None or {} or ''}



import asyncio
from uuid import UUID
import logging
from . import context


def _check_token(token: str) -> bool:
    try:
        UUID(token, version=4)
        return True
    except ValueError:
        return False


def _setup_callbacks(executor, on_startup=None, on_shutdown=None):
    if on_startup is not None:
        executor.on_startup(on_startup)
    if on_shutdown is not None:
        executor.on_shutdown(on_shutdown)


def start_polling(dispatcher, *, loop=None, skip_updates=False, timeout=None,
                  on_startup=None, on_shutdown=None):
    if not _check_token(dispatcher.bot.token):
        raise Exception('Token ' + dispatcher.bot.token + ' is not valid')
    executor = Executor(dispatcher, loop=loop)
    _setup_callbacks(executor, on_startup, on_shutdown)
    executor.start_polling(timeout=timeout)


class Executor:
    def __init__(self, dispatcher, skip_updates=None, loop=None):
        if loop is None:
            loop = dispatcher.loop
        self.loop = loop
        self.dispatcher = dispatcher
        self.skip_updates = skip_updates
        self._freeze = False

        self._on_startup_polling = []

    @property
    def frozen(self):
        return self._freeze

    def on_startup(self, callback: callable, polling=True, webhook=False):
        logger = logging.getLogger('Executor(on_startup)')
        logger.info('on START')

    def on_shutdown(self, callback: callable, polling=True, webhook=False):
        logger = logging.getLogger('Executor(on_shutdown)')
        logger.info('on SHUT')

    def _check_frozen(self):
        if self.frozen:
            raise RuntimeError('Executor is frozen')

    def _prepare_polling(self):
        self._check_frozen()
        self._freeze = True

        self.loop.set_task_factory(context.task_factory)

    def start_polling(self, timeout=None):

        self._prepare_polling()
        loop: asyncio.AbstractEventLoop = self.loop

        try:
            loop.run_until_complete(self._startup_polling())
            loop.create_task(self.dispatcher.start_polling())
            loop.run_forever()
        except (KeyboardInterrupt, SystemExit):
            pass
        finally:
            loop.run_until_complete(self._shutdown_polling())

    async def _skip_updates(self):
        print('skip updates')

    async def _startup_polling(self):
        if self.skip_updates:
            await self._skip_updates()
        for callback in self._on_startup_polling:
            await callback(self.dispatcher)

    async def _shutdown_polling(self):
        await self._shutdown()

    async def _shutdown(self):
        await self.dispatcher.stop_polling()



from pyAitu.models.constants.file_type import FileType
from pyAitu.utils.strings import FILE_ID, FILE_TYPE, NAME


class Media:
    def __init__(self, file_id, name, file_type: FileType):
        self._file_id = file_id
        self._name = name
        self._file_type = file_type

    def to_dict(self):
        return {
            FILE_ID: self._file_id,
            NAME: self._name,
            FILE_TYPE: self._file_type
        }



import functools
from ..peer import Peer, PeerUser
from ..media import Image, Audio
from ...utils import helper


class Message:
    def __init__(self, json_object):
        self.updateId = json_object.get("updateId")
        self.messageId = json_object.get("messageId")
        self.dialog = Peer(json_object.get("dialog"))
        self.content = json_object.get("content")
        if json_object.get("author").get("type") == 'USER':
            self.author = PeerUser(json_object.get("author"))
        self.media = []

        if json_object.get("media"):
            for item in json_object.get("media"):
                if item.get("type") == "Image":
                    self.media.append(Image(item))
                elif item.get("type") == "Audio":
                    self.media.append(Audio(item))

    def is_command(self):
        return self.content and self.content.startswith('/')

    @property
    def chat(self):
        return self.dialog

    @property
    def text(self):
        return self.content

    @property
    @functools.lru_cache()
    def content_type(self):
        if self.media and self.media[0].type == "Image":
            return ContentType.PHOTO[0]
        elif self.media and self.media[0].type == "Audio":
            return ContentType.AUDIO[0]
        else:
            return ContentType.TEXT[0]


class ContentType(helper.Helper):
    mode = helper.HelperMode.snake_case

    TEXT = helper.ListItem()
    PHOTO = helper.ListItem()
    VIDEO = helper.ListItem()
    AUDIO = helper.ListItem()
    ANY = helper.ListItem()




from typing import Iterable


class MiddlewareManager:
    def __init__(self, dispatcher):
        self.dispatcher = dispatcher
        self.loop = dispatcher.loop
        self.bot = dispatcher.bot
        self.applications = []

    async def trigger(self, action: str, args: Iterable):
        for app in self.applications:
            await app.trigger(action,args)


from .content import Content
from ..form_action import FormAction
from ..options import Options


class Submit(Content):
    def __init__(self,
                 content_id: str,
                 title: str,
                 form_action: FormAction,
                 options: Options = None):
        super().__init__(content_type="submit", content_id=content_id)
        self.title = title
        self.form_action = form_action.__dict__
        self.options = options.__dict__ if options is not None else {}



from ..options import Options
from .content import Content


class UserInfo(Content):
    def __init__(
            self,
            content_id: str,
            user_id: str,
            options: Options = None
    ):
        super().__init__(content_type="user_info", content_id=content_id)
        self.user_id = user_id
        if options is not None:
            self.options = options



class Indent:
    def __init__(self, left: int = None, top: int = None, right: int = None, bottom: int = None):
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom



import asyncio
import logging
from typing import Optional
from .storage import BaseStorage, DisabledStorage
from .handler import Handler
from .middlewares import MiddlewareManager
from .filters import generate_default_filters
from ..models import ContentType
from ..bot import Bot
from ..utils import context

log = logging.getLogger(__name__)

MODE = 'MODE'
LONG_POLLING = 'long-polling'
UPDATE_OBJECT = 'update_object'


class Dispatcher:
    def __init__(self, bot, loop=None, storage: Optional[BaseStorage] = None, run_tasks_by_default: bool = False):
        if loop is None:
            loop = bot.loop
        if storage is None:
            storage = DisabledStorage()
        self.bot: Bot = bot
        self.loop = loop
        self.storage = storage
        self._polling = False
        self.run_tasks_by_default = run_tasks_by_default
        self._close_waiter = loop.create_future()

        self.updates_handler = Handler(self, middleware_key='update')
        self.message_handlers = Handler(self, middleware_key='message')
        self.quick_button_handlers = Handler(self, middleware_key='quick_button')
        self.message_id_assigned_handlers = Handler(self, middleware_key='message_id_assigned')
        self.inline_command_handlers = Handler(self, middleware_key='inline_command')
        self.form_closed_handlers = Handler(self, middleware_key='form_closed')
        self.form_submitted_handlers = Handler(self, middleware_key='form_submitted')
        self.form_message_sent_handlers = Handler(self, middleware_key='form_message_sent')
        self.error_handlers = Handler(self, once=False, middleware_key='error')
        self.middleware = MiddlewareManager(self)

        self.updates_handler.register(self.process_update)

    async def start_polling(self):
        if self._polling:
            raise RuntimeError('Polling already started')

        log.info('Polling started')

        context.set_value(MODE, LONG_POLLING)
        context.set_value('dispatcher', self)
        context.set_value('bot', self.bot)

        self._polling = True

        try:
            while self._polling:
                try:
                    updates = await self.bot.get_updates()
                except:
                    log.exception('Cause exception while getting updates')
                    await asyncio.sleep(15)
                    continue

                if updates:
                    log.debug(f"Received {len(updates)} updates.")
                    self.loop.create_task(self._process_polling_updates(updates))
        finally:
            self._close_waiter.set_result(None)
            log.warning('Polling is stopped')

    async def _process_polling_updates(self, updates):
        await self.process_updates(updates)

    async def process_updates(self, updates):
        tasks = []
        for update in updates:
            tasks.append(self.updates_handler.notify(update))
        return await asyncio.gather(*tasks)

    async def stop_polling(self):
        if self._polling:
            log.info('Stop polling...')
            self._polling = False

    def message_handler(self, *custom_filters, commands=None, content_types=None, state=None, run_task=None, **kwargs):
        def decorator(callback):
            self.register_message_handler(callback,
                                          commands=commands,
                                          content_types=content_types,
                                          custom_filters=custom_filters,
                                          state=state,
                                          run_task=run_task,
                                          **kwargs)
            return callback

        return decorator

    def register_message_handler(self,
                                 callback, *,
                                 commands=None,
                                 content_types=None,
                                 custom_filters=None,
                                 state=None,
                                 run_task=None,
                                 **kwargs):

        if content_types is None:
            content_types = ContentType.TEXT
        if custom_filters is None:
            custom_filters = []

        filters_set = generate_default_filters(self,
                                               *custom_filters,
                                               commands=commands,
                                               state=state,
                                               content_types=content_types,
                                               **kwargs)
        self.message_handlers.register(self._wrap_async_task(callback, run_task), filters_set)

    def quick_button_handler(self, *custom_filters, func=None, state=None, run_task=None, **kwargs):
        def decorator(callback):
            self.register_quick_button_handler(callback, func=func, state=state,
                                               custom_filters=custom_filters, run_task=run_task, **kwargs)
            return callback

        return decorator

    def message_id_assigned_handler(self, *custom_filters, func=None, state=None, run_task=None, **kwargs):
        def decorator(callback):
            self.register_message_id_assigned_handler(callback, func=func, state=state,
                                                      custom_filters=custom_filters, run_task=run_task, **kwargs)
            return callback

        return decorator

    def register_message_id_assigned_handler(self, callback, *, func=None,
                                             state=None, custom_filters=None, run_task=None, **kwargs):
        if custom_filters is None:
            custom_filters = []

        filters_set = generate_default_filters(self,
                                               *custom_filters,
                                               func=func,
                                               state=state,
                                               **kwargs)
        self.message_id_assigned_handlers.register(self._wrap_async_task(callback, run_task), filters_set)

    def register_quick_button_handler(self, callback, *, func=None,
                                      state=None, custom_filters=None, run_task=None, **kwargs):
        if custom_filters is None:
            custom_filters = []

        filters_set = generate_default_filters(self,
                                               *custom_filters,
                                               func=func,
                                               state=state,
                                               **kwargs)
        self.quick_button_handlers.register(self._wrap_async_task(callback, run_task), filters_set)

    def inline_command_handler(self, *custom_filters, func=None, state=None, run_task=None, **kwargs):
        def decorator(callback):
            self.register_inline_command_handler(callback, func=func, state=state,
                                                 custom_filters=custom_filters, run_task=run_task, **kwargs)
            return callback

        return decorator

    def register_inline_command_handler(self, callback, *, func=None, state=None,
                                        custom_filters=None, run_task=None, **kwargs):
        if custom_filters is None:
            custom_filters = []
        filter_set = generate_default_filters(self,
                                              *custom_filters,
                                              func=func,
                                              state=state,
                                              **kwargs)
        self.inline_command_handlers.register(self._wrap_async_task(callback, run_task), filter_set)

    def form_closed_handler(self, *custom_filters, func=None, state=None, run_task=None, **kwargs):
        def decorator(callback):
            self.register_form_closed_handler(callback, *custom_filters, func=func,
                                              state=state, run_task=run_task, **kwargs)
            return callback

        return decorator

    def register_form_closed_handler(self, callback, *, func=None, state=None,
                                     custom_filters=None, run_task=None, **kwargs):
        if custom_filters is None:
            custom_filters = []
        filter_set = generate_default_filters(self,
                                              *custom_filters,
                                              func=func,
                                              state=state,
                                              **kwargs)
        self.form_closed_handlers.register(self._wrap_async_task(callback, run_task), filter_set)

    def form_submitted_handler(self, *custom_filters, func=None, state=None, run_task=None, **kwargs):
        def decorator(callback):
            self.register_form_submitted_handler(callback, *custom_filters, func=func,
                                                 state=state, run_task=run_task, **kwargs)
            return callback

        return decorator

    def register_form_submitted_handler(self, callback, *, func=None, state=None,
                                        custom_filters=None, run_task=None, **kwargs):
        if custom_filters is None:
            custom_filters = []
        filter_set = generate_default_filters(self,
                                              *custom_filters,
                                              func=func,
                                              state=state,
                                              **kwargs)
        self.form_submitted_handlers.register(self._wrap_async_task(callback, run_task), filter_set)

    def form_message_sent_handler(self, *custom_filters, func=None, state=None, run_task=None, **kwargs):
        def decorator(callback):
            self.register_form_message_sent_handler(callback, *custom_filters, func=func,
                                                    state=state, run_task=run_task, **kwargs)
            return callback

        return decorator

    def register_form_message_sent_handler(self, callback, *, func=None, state=None,
                                           custom_filters=None, run_task=None, **kwargs):
        if custom_filters is None:
            custom_filters = []
        filter_set = generate_default_filters(self,
                                              *custom_filters,
                                              func=func,
                                              state=state,
                                              **kwargs)
        self.form_message_sent_handlers.register(self._wrap_async_task(callback, run_task), filter_set)

    def _wrap_async_task(self, callback, run_task=None) -> callable:
        if run_task is None:
            run_task = self.run_tasks_by_default

        if run_task:
            return self.async_task(callback)
        return callback

    def async_task(self, callback):
        print('async task called')

    async def process_update(self, update):
        context.set_value(UPDATE_OBJECT, update)

        try:
            if update.message:
                await self.state_updater(update.message)
                return await self.message_handlers.notify(update.message)
            elif update.quick_button_selected:
                await self.state_updater(update.quick_button_selected)
                return await self.quick_button_handlers.notify(update.quick_button_selected)
            elif update.inline_command_selected:
                await self.state_updater(update.inline_command_selected)
                return await self.inline_command_handlers.notify(update.inline_command_selected)
            elif update.form_closed:
                await self.state_updater(update.form_closed)
                return await self.form_closed_handlers.notify(update.form_closed)
            elif update.form_message_sent:
                await self.state_updater(update.form_message_sent)
                return await self.form_message_sent_handlers.notify(update.form_message_sent)
            elif update.form_submitted:
                await self.state_updater(update.form_submitted)
                return await self.form_submitted_handlers.notify(update.form_submitted)
            elif update.message_id_assigned:
                await self.state_updater(update.message_id_assigned)
                return await self.message_id_assigned_handlers.notify(update.message_id_assigned)
            else:
                print("Undefined update " + update.updateType)
        except Exception as e:
            err = await self.error_handlers.notify(self, update, e)
            if err:
                return err
            raise

    async def state_updater(self, update):
        state = await self.storage.get_state(
            chat=update.dialog.id,
            user=update.sender.id if hasattr(update, "sender") else update.dialog.id
        )
        context.update_state(
            user=update.sender.id if hasattr(update, "sender") else update.dialog.id,
            state=state
        )



from typing import Optional, List, Dict, Any
from pyAitu.utils.dictionary_extractor import dictionary_of_object_if_exist

# Type aliases
Object = Any
Array = List
Objects = Array[Object]
Dictionaries = Array[Dict[Any, Any]]


def serialized(objects: Optional[Objects]) -> Optional[Dictionaries]:
    if not isinstance(objects, Array) or objects is None:
        return None
    objects = list(map(lambda object: dictionary_of_object_if_exist(object), objects))
    objects = list(filter(None, objects))
    return objects



from .content import Content
from .file_metadata import FileMetadata
from ..options import Options
from ..form_action import FormAction


class Image(Content):
    def __init__(self,
                 content_id: str,
                 file_metadata: FileMetadata,
                 options: Options = None,
                 form_action: FormAction = None):
        super().__init__(content_type="image", content_id=content_id)
        self.file_metadata = file_metadata.__dict__
        self.options = options.__dict__ if options is not None else {}
        self.form_action = form_action.__dict__ if form_action is not None else None



from typing import List, Optional
from ..options import Options
from .content import Content
from ..validation_rule import ValidationRule
from .item import Item
from pyAitu.utils.serializer import serialized
from pyAitu.utils.dictionary_extractor import dictionary_of_object_if_exist


class CustomContainer(Content):
    def __init__(
            self,
            content_id: str,
            content: List[Content],
            options: Optional[Options] = None

    ):
        super().__init__(content_type="custom_container", content_id=content_id)
        self.content = serialized(content)
        self.options = dictionary_of_object_if_exist(options)



import logging
import aiohttp
import json
import os
from http import HTTPStatus

log = logging.getLogger('pyAitu')

API_URL = "https://messapi.btsdapps.net/bot/v1/updates/"
FILE_UPLOAD_URL = "https://messapi.btsdapps.net/bot/v1/upload/"
FILE_DOWNLOAD_URL = "https://messapi.btsdapps.net/bot/v1/download/"
WEBHOOK_INFO_URL = "https://messapi.btsdapps.net/bot/v1/webhook/"


def _compose_data(params=None, files=None):
    data = aiohttp.formdata.FormData(quote_fields=False)
    if params:
        for key, value in params.items():
            data.add_field(key, str(value))

    if files:
        for key, f in files.items():
            data.add_field('key', open(f, 'rb'), filename=os.path.basename(f), content_type='application/octet-stream')

    return data


async def request(session, token, method, data=None, files=None, **kwargs) -> bool or dict:
    log.debug(f"Making request {method}")

    if files and method == "UploadFile":
        data = _compose_data(data, files)
        try:
            async with session.post(FILE_UPLOAD_URL, data=data, headers={"X-BOT-TOKEN": token}, **kwargs) as response:
                return await _check_result(method, response)
        except aiohttp.ClientError as e:
            raise Exception(f"aiohttp client throws an error: {e.__class__.__name__}: {e}")
    elif method in ["SetWebhook", "GetWebhook", "DeleteWebhook"]:
        if method == "GetWebhook":
            try:
                async with session.get(WEBHOOK_INFO_URL, headers={"X-BOT-TOKEN": token}, **kwargs) as response:
                    return await _check_result(method, response)
            except aiohttp.ClientError as e:
                raise Exception(f"aiohttp client throws an error: {e.__class__.__name__}: {e}")
        elif method == "SetWebhook":
            try:
                async with session.post(WEBHOOK_INFO_URL, json=data, headers={"X-BOT-TOKEN": token}, **kwargs) \
                        as response:
                    return await _check_result(method, response)
            except aiohttp.ClientError as e:
                raise Exception(f"aiohttp client throws an error: {e.__class__.__name__}: {e}")
        elif method == "DeleteWebhook":
            try:
                async with session.delete(WEBHOOK_INFO_URL, headers={"X-BOT-TOKEN": token}, **kwargs) as response:
                    return await _check_result(method, response)
            except aiohttp.ClientError as e:
                raise Exception(f"aiohttp client throws an error: {e.__class__.__name__}: {e}")
    else:
        try:
            if method == "GetUpdates":
                async with session.get(API_URL, headers={"X-BOT-TOKEN": token}, **kwargs) as response:
                    return await _check_result(method, response)
            else:
                async with session.post(API_URL, json=data, headers={"X-BOT-TOKEN": token}, **kwargs) as response:
                    return await _check_result(method, response)
        except aiohttp.ClientError as e:
            raise Exception(f"aiohttp client throws an error: {e.__class__.__name__}: {e}")


async def _check_result(method_name, response):
    body = await response.text()
    log.debug(f"Response for {method_name}: [{response.status}] {body}")

    if response.content_type != 'application/json':
        raise Exception(f"Invalid response with content type {response.content_type}: \"{body}\"")

    try:
        result_json = await response.json(loads=json.loads)
    except ValueError:
        result_json = {}

    if result_json.get('error') is not None:
        error = str(result_json.get('error'))
        message = str(result_json.get('message'))
        status = response.status
        log.error(f' - {status}: {error} - {message}')

    description = result_json.get('description') or body
    if HTTPStatus.OK <= response.status <= HTTPStatus.IM_USED:
        return result_json



import pathlib
import re
import sys
from setuptools import setup, find_packages

WORK_DIR = pathlib.Path(__file__).parent

MINIMAL_PY_VERSION = (3, 7)
if sys.version_info < MINIMAL_PY_VERSION:
    raise RuntimeError('pyAitu works only with Python {}+'.format('.'.join(map(str, MINIMAL_PY_VERSION))))


def get_description():
    with open('README.rst', 'r', encoding='utf-8') as f:
        return f.read()


def get_version():
    txt = (WORK_DIR / 'pyAitu' / '__init__.py').read_text('utf_8')
    try:
        return re.findall(r"^__version__ = '([^']+)'\r?$", txt, re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')


REQUIRED_PACKAGES = [
    'aiohttp>=3.5.4, <4.0.0',
    'certifi>=2018.8.24'
]


setup(
    name="pyAitu",
    version="0.0.3",
    author="Yerassyl Zeinolla",
    author_email="yerassyl.zeinolla@btsdigital.kz",
    description="Asynchronous Python framework for Aitu Bot API",
    long_description=get_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/btsdigital/pyAitu",
    packages=find_packages(exclude=('tests', 'tests.*', 'examples.*', 'docs')),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
    ],
    python_requires='>=3.7',
    install_requires=REQUIRED_PACKAGES,
    package_data={'': ['requirements.txt']},
    include_package_data=False,
    license='Apache 2.0'
)



from .content import Content
from ..options import Options
from ..validation_rule import ValidationRule
from typing import Optional, List
from pyAitu.utils.serializer import serialized
from pyAitu.utils.dictionary_extractor import dictionary_of_object_if_exist

# Type alias
ValidationRules = List[ValidationRule]


class Input(Content):
    def __init__(
            self,
            content_id: str,
            placeholder: Optional[str] = None,
            title: Optional[str] = None,
            mask: Optional[str] = None,
            text: Optional[str] = None,
            options: Optional[Options] = None,
            validation_rules: Optional[ValidationRules] = None,
    ):
        super().__init__(content_type="input", content_id=content_id)
        self.placeholder = placeholder
        self.title = title
        self.mask = mask
        self.text = text
        self.options = dictionary_of_object_if_exist(options)
        self.validations_rules = serialized(validation_rules)



from .options import Options


class Header:
    def __init__(self, _type: str, title: str, options: Options = None):
        self.type = _type
        self.title = title
        self.options = options.__dict__ if options is not None else {}



class FormAction:
    def __init__(
            self,
            action: str,
            data_template: str = None,
            hidden_metadata: str = None
    ):
        self.action = action
        self.data_template = data_template
        self.hidden_metadata = hidden_metadata



class Image:
    def __init__(self, json_object):
        self.type = json_object.get("type")
        self.fileId = json_object.get("fileId")
        self.mimeType = json_object.get("mimeType")
        self.name = json_object.get("name")
        self.size = json_object.get("size")
        self.width = json_object.get("width")
        self.height = json_object.get("height")


from ...utils.strings import DEFAULT_USER, TYPE, ID


class Recipient:
    def __init__(self, _id: str):
        self.default_user = DEFAULT_USER
        self._id = _id

    def get_recipient(self):
        return {
            TYPE: self.default_user,
            ID: self._id
        }



from typing import List, Optional

from .content import Content
from ..validation_rule import ValidationRule
from ..options import Options
from ....utils.serializer import serialized

ValidationRules = List[ValidationRule]


class MediaPicker(Content):
    def __init__(
            self,
            content_id: str,
            title: str,
            options: Options,
            validations_rules: Optional[ValidationRules] = None,
    ):
        super().__init__(content_id=content_id, content_type="media_picker")
        self.title = title
        self.options = options.__dict__
        self.validations_rules = serialized(validations_rules)



from .inline_command import InlineCommand
from .quick_button_command import QuickButtonCommand
from .reply_command import ReplyCommand
from .command import Command
from .uiState import UiState
from .form_message import FormMessage

__all__ = [
    InlineCommand,
    QuickButtonCommand,
    ReplyCommand,
    Command,
    UiState,
    FormMessage
]



from typing import List, Optional
from ..options import Options
from .content import Content
from ..validation_rule import ValidationRule
from .item import Item
from pyAitu.utils.serializer import serialized
from pyAitu.utils.dictionary_extractor import dictionary_of_object_if_exist


class RadioItem:
    def __init__(self, content_id: str, title: str):
        self.content_id = content_id
        self.title = title


class Radiogroup(Content):
    def __init__(
            self,
            content_id: str,
            title: Optional[str] = None,
            items: List[Item] = None,
            validations_rules: Optional[List[ValidationRule]] = None,
            options: Optional[Options] = None,
            default_value: Optional[Item] = None
    ):
        super().__init__(content_type="radiogroup", content_id=content_id)
        self.title = title
        self.items = serialized(items)
        self.validations_rules = serialized(validations_rules)
        self.options = dictionary_of_object_if_exist(options)
        self.default_value = serialized(default_value)
